
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbShapes
#define HDR_dbShapes

#include <dbObject.h>
#include <dbManager.h>
#include <dbLayer.h>
#include <dbLayoutStateModel.h>
#include <tlVector.h>

namespace db 
{

template <class Shapes, class Sh> class layer_op;
template <class Coord> class shapes;
template <class Coord> class shape_iterator;
template <class Coord> class generic_polygon_edge_iterator;
template <class Coord> class generic_repository;

/**
 *  @brief A base class for layer operations 
 *
 *  This class is used for the Op classes for the undo/redo queuing mechanism.
 */
template <class Shapes>
class layer_op_base
  : public db::Op
{
public:
  layer_op_base () : db::Op () { }

  virtual void undo (Shapes *shapes) = 0;
  virtual void redo (Shapes *shapes) = 0;
};

/**
 *  @brief A helper class for shape generalisation
 *
 *  This class serves first as a RTTI token for the 
 *  various shape-specific layer implementations and 
 *  provides some common methods though it's interface
 */

template <class Shapes>
class layer_base 
{
public:
  typedef typename Shapes::box_type box_type;
  typedef typename Shapes::coord_type coord_type;

  virtual ~layer_base () { }

  virtual box_type bbox () const = 0;
  virtual void update_bbox () = 0;
  virtual bool is_bbox_dirty () const = 0;
  virtual void sort () = 0;
  virtual void clear (Shapes *target, db::Manager *manager) = 0;
  virtual layer_base<Shapes> *clone (Shapes *target, db::Manager *manager) const = 0;
  virtual layer_base<Shapes> *translated (Shapes *target, generic_repository<coord_type> &rep) const = 0;

  virtual void collect_mem_stat (db::MemStatistics &m) const = 0;
};

/**
 *  @brief Actual implementation of the layer_base class
 *  
 *  For a certain shape type this template implements the 
 *  shape specific layer methods.
 */

template <class Shapes, class Sh>
class layer_class 
  : public layer_base<Shapes>
{
public:
  typedef db::layer<Sh> layer_type;
  typedef typename Shapes::box_type box_type;
  typedef typename Shapes::coord_type coord_type;

  layer_class ()
    : layer_base<Shapes> (), m_layer () 
  { }

  layer_type &layer () 
  {
    return m_layer;
  }

  const layer_type &layer () const
  {
    return m_layer;
  }

  virtual box_type bbox () const 
  {
    return m_layer.bbox ();
  }

  virtual void update_bbox () 
  {
    m_layer.update_bbox ();
  }

  virtual bool is_bbox_dirty () const
  {
    return m_layer.is_bbox_dirty ();
  }

  virtual void sort () 
  {
    m_layer.sort ();
  }

  virtual void clear (Shapes *target, db::Manager *manager)
  {
    if (manager && manager->transacting ()) {
      manager->queue (target, new db::layer_op<Shapes, Sh> (false /*not insert*/, m_layer.begin (), m_layer.end ()));
    }
    m_layer.clear ();
  }

  virtual layer_base<Shapes> *clone (Shapes *target, db::Manager *manager) const 
  {
    layer_class<Shapes, Sh> *r = new layer_class<Shapes, Sh> ();
    if (manager && manager->transacting ()) {
      manager->queue (target, new db::layer_op<Shapes, Sh> (true /*insert*/, m_layer.begin (), m_layer.end ()));
    }
    r->m_layer = m_layer;
    return r;
  }
  
  virtual layer_base<Shapes> *translated (Shapes *target, generic_repository<coord_type> &rep) const 
  {
    layer_class<Shapes, Sh> *r = new layer_class<Shapes, Sh> ();
    r->m_layer.translate (m_layer, rep);
    return r;
  }
  
  virtual void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.shapes_info ((const layer_base<Shapes> &)(*this)); // base class contribution
    m_layer.collect_mem_stat (m);
  }

private:
  layer_type m_layer;
};

/**
 *  @brief A helper compare functor for array member Shape's and a displacement
 */
template <class Sh>
struct compare_array_displacement
{
  typedef typename Sh::disp_type disp_type;
  bool operator() (const Sh &shape, const disp_type &disp) const
  {
    return shape.array_trans ().disp () < disp.disp ();
  }
  bool operator() (const disp_type &disp, const Sh &shape) const
  {
    return disp.disp () < shape.array_trans ().disp ();
  }
};

/**
 *  @brief A "shapes" collection
 *  
 *  A shapes collection is a collection of geometrical objects.
 *  The implementation is based on a set of layers of different
 *  shape types. 
 *  The general idea is that is rarely required to operate with
 *  different shape types at once. Therefore it is practical to
 *  request only a single shape type at once. This is done through
 *  the various begin.. and end.. methods that are specialized on
 *  a certain shape type.
 */

template <class Coord>
class shapes 
  : public db::Object
{
public:
  typedef Coord coord_type;
  typedef db::box<Coord> box_type;
  typedef db::shape_iterator<Coord> shape_iterator;
  typedef typename shape_iterator::shape_type shape_type;
  typedef db::generic_polygon_edge_iterator<Coord> polygon_edge_iterator;

  /**
   *  @brief Standard ctor: create an empty collection referencing a graph
   *
   *  The graph reference is used to invalid the bbox flag of the graph
   *  whenever something changes on the shapes list.
   */
  shapes (db::Manager *manager = 0, db::LayoutStateModel *state_model = 0) 
    : db::Object (manager), 
      mp_state_model (state_model)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Dtor: clear all ..
   */
  ~shapes () 
  {
    clear ();
    mp_state_model = 0;
  }

  /**
   *  @brief Copy ctor
   */
  shapes (const shapes &d) 
    : db::Object (d), mp_state_model (d.mp_state_model)
  {
    operator= (d);
  }

  /**
   *  @brief Assignment operator
   */
  shapes &
  operator= (const shapes &d)
  {
    if (&d != this) {

      clear ();

      m_layers.reserve (d.m_layers.size ());
      for (typename tl::vector<layer_base<shapes> *>::const_iterator l = d.m_layers.begin (); l != d.m_layers.end (); ++l) {
        m_layers.push_back ((*l)->clone (this, manager ()));
      }

    }
    return *this;
  }

  /**
   *  @brief The translation operator
   *
   *  This operator is used to copy one shapes layer to another repository space.
   *  The current shapes layer will be overwritten. 
   *
   *  @param src The source shapes collection
   *  @param rep The repository that is associated with *this and into which the
   *             shapes will be copied
   */
  void translate (const shapes<Coord> &d, generic_repository<Coord> &rep)
  {
    tl_assert (&d != this);
    //  no undo support for this currently
    tl_assert (! manager () || ! manager ()->transacting ());

    clear ();

    m_layers.reserve (d.m_layers.size ());
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = d.m_layers.begin (); l != d.m_layers.end (); ++l) {
      m_layers.push_back ((*l)->translated (this, rep));
    }
  }

  /**
   *  @brief Swap the contents of this shapes collection with another one
   */
  void swap (shapes &d)
  {
    tl_assert (! manager () || ! manager ()->transacting ()); // TODO: implement undo/redo support?
    d.invalidate_state ();  //  HINT: must come before the change is done!
    invalidate_state ();
    m_layers.swap (d.m_layers);
  }

  /**
   *  @brief Insert a shape of the given type
   *
   *  This inserts a shape of the given kind into the collection.
   *  Basically a new layer is created if no shape of this kind
   *  already existed.
   *  Inserting a shape will invalidate the bbox and the sorting
   *  state.
   *
   *  @param sh The shape to insert (copy)
   *  
   *  @return A reference to the object created
   */
  template <class Sh>
  const Sh &insert (const Sh &sh)
  {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, Sh> (true /*insert*/, sh));
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    return layer<Sh> ().insert (sh);
  }

  /**
   *  @brief Insert a shape sequence
   *
   *  Inserts a sequence of shapes [from,to)
   */

  template <class Iter>
  void insert (Iter from, Iter to)
  {
    typedef typename std::iterator_traits <Iter>::value_type value_type;
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, value_type> (true /*insert*/, from, to));
    }
    invalidate_state ();
    layer<value_type> ().insert (from, to);
  }

  /**
   *  @brief Reserve the number of elements for a shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param n The number of elements to reserve
   */
  template <class Tag> 
  void reserve (Tag tag, size_t n)
  {
    layer<typename Tag::shape_type> ().reserve (n);
  }

  /**
   *  @brief Erase an element 
   *
   *  Erases a shape at the given position
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param pos The position of the shape to erase
   */
  template <class Tag> 
  void erase (Tag tag, typename db::layer<typename Tag::shape_type>::iterator pos)
  {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, *pos));
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    layer<typename Tag::shape_type> ().erase (pos);
  }

  /**
   *  @brief Erase an element by the shape reference
   *
   *  Given a shape reference, the corresponding shape is erased.
   *  It can be specified, whether the whole array is deleted or just the instance 
   *  if the shape references an array member.
   *  WARNING: After the shape is deleted, other shape references may become invalid.
   *
   *  @param shape The reference to the shape to delete
   *  @param whole_array Delete the array as a whole, if true
   */
  void erase_shape (const shape_type &shape, bool whole_array)
  {
    switch (shape.m_type) {
    case shape_type::Null:
      break;
    case shape_type::Polygon:
      erase_shape_by_tag (typename shape_type::polygon_type::tag (), shape);
      break;
    case shape_type::PolygonRef:
      erase_shape_by_tag (typename shape_type::polygon_ref_type::tag (), shape);
      break;
    case shape_type::PolygonPtrArrayMember:
      if (! whole_array) { 
        resolve_array_except (typename shape_type::polygon_ptr_array_type::tag (), typename shape_type::polygon_ref_type::tag (), shape);
        break;
      }
    case shape_type::PolygonPtrArray:
      erase_shape_by_tag (typename shape_type::polygon_ptr_array_type::tag (), shape);
      break;
    case shape_type::SimplePolygon:
      erase_shape_by_tag (typename shape_type::simple_polygon_type::tag (), shape);
      break;
    case shape_type::SimplePolygonRef:
      erase_shape_by_tag (typename shape_type::simple_polygon_ref_type::tag (), shape);
      break;
    case shape_type::SimplePolygonPtrArrayMember:
      if (! whole_array) { 
        resolve_array_except (typename shape_type::simple_polygon_ptr_array_type::tag (), typename shape_type::simple_polygon_ref_type::tag (), shape);
        break;
      }
    case shape_type::SimplePolygonPtrArray:
      erase_shape_by_tag (typename shape_type::simple_polygon_ptr_array_type::tag (), shape);
      break;
    case shape_type::Edge:
      erase_shape_by_tag (typename shape_type::edge_type::tag (), shape);
      break;
    case shape_type::Path:
      erase_shape_by_tag (typename shape_type::path_type::tag (), shape);
      break;
    case shape_type::PathRef:
      erase_shape_by_tag (typename shape_type::path_ref_type::tag (), shape);
      break;
    case shape_type::PathPtrArrayMember:
      if (! whole_array) { 
        resolve_array_except (typename shape_type::path_ptr_array_type::tag (), typename shape_type::path_ref_type::tag (), shape);
        break;
      }
    case shape_type::PathPtrArray:
      erase_shape_by_tag (typename shape_type::path_ptr_array_type::tag (), shape);
      break;
    case shape_type::Box:
      erase_shape_by_tag (typename shape_type::box_type::tag (), shape);
      break;
    case shape_type::BoxArrayMember:
      if (! whole_array) { 
        resolve_box_array_except (typename shape_type::box_array_type::tag (), typename shape_type::box_type::tag (), shape);
        break;
      }
    case shape_type::BoxArray:
      erase_shape_by_tag (typename shape_type::box_array_type::tag (), shape);
      break;
    case shape_type::ShortBox:
      erase_shape_by_tag (typename shape_type::short_box_type::tag (), shape);
      break;
    case shape_type::ShortBoxArrayMember:
      if (! whole_array) { 
        resolve_box_array_except (typename shape_type::short_box_array_type::tag (), typename shape_type::short_box_type::tag (), shape);
        break;
      }
    case shape_type::ShortBoxArray:
      erase_shape_by_tag (typename shape_type::short_box_array_type::tag (), shape);
      break;
    case shape_type::Text:
      erase_shape_by_tag (typename shape_type::text_type::tag (), shape);
      break;
    case shape_type::TextRef:
      erase_shape_by_tag (typename shape_type::text_ref_type::tag (), shape);
      break;
    case shape_type::TextPtrArrayMember:
      if (! whole_array) { 
        resolve_array_except (typename shape_type::text_ptr_array_type::tag (), typename shape_type::text_ref_type::tag (), shape);
        break;
      }
    case shape_type::TextPtrArray:
      erase_shape_by_tag (typename shape_type::text_ptr_array_type::tag (), shape);
      break;
    case shape_type::UserObject:
      erase_shape_by_tag (typename shape_type::user_object_type::tag (), shape);
      break;
    };
  }

  /**
   *  @brief Erase an element by the shape reference
   *
   *  Given a set of ordered shape reference, the corresponding shapes are erased.
   *  It can be specified, whether the whole array is deleted or just the instance 
   *  if the shape references an array member.
   *  Hint: for the following code it is important that the ArrayMember types come *after*
   *  the non-member types. This is because this way it is ensured that array members get erased
   *  after the respective single shapes/shape instances. That guarantees that if single shapes
   *  are created on array resolution these shapes can be added to the single shape lists without
   *  invalidating other shape references into them in db::shapes::erase with multi-shape erase.
   *  WARNING: After the shapes are deleted, other shape references may become invalid.
   *
   *  @param shapes The reference to the shapes to delete. This array must be sorted using the Shape's operator<
   *  @param whole_array Delete the arrays as a whole, if true
   */
  void erase_shapes (const std::vector<shape_type> &shapes, bool whole_array)
  {
    for (typename std::vector<shape_type>::const_iterator s = shapes.begin (); s != shapes.end (); ) {

      typename std::vector<shape_type>::const_iterator snext = s;
      while (snext != shapes.end () && snext->type () == s->type () && snext->has_prop_id () == s->has_prop_id ()) {
        ++snext;
      }

      switch (s->m_type) {
      case shape_type::Null:
        break;
      case shape_type::Polygon:
        erase_shapes_by_tag (typename shape_type::polygon_type::tag (), s, snext);
        break;
      case shape_type::PolygonRef:
        erase_shapes_by_tag (typename shape_type::polygon_ref_type::tag (), s, snext);
        break;
      case shape_type::PolygonPtrArrayMember:
        if (! whole_array) { 
          resolve_array_except (typename shape_type::polygon_ptr_array_type::tag (), typename shape_type::polygon_ref_type::tag (), s, snext);
          break;
        }
      case shape_type::PolygonPtrArray:
        erase_shapes_by_tag (typename shape_type::polygon_ptr_array_type::tag (), s, snext);
        break;
      case shape_type::SimplePolygon:
        erase_shapes_by_tag (typename shape_type::simple_polygon_type::tag (), s, snext);
        break;
      case shape_type::SimplePolygonRef:
        erase_shapes_by_tag (typename shape_type::simple_polygon_ref_type::tag (), s, snext);
        break;
      case shape_type::SimplePolygonPtrArrayMember:
        if (! whole_array) { 
          resolve_array_except (typename shape_type::simple_polygon_ptr_array_type::tag (), typename shape_type::simple_polygon_ref_type::tag (), s, snext);
          break;
        }
      case shape_type::SimplePolygonPtrArray:
        erase_shapes_by_tag (typename shape_type::simple_polygon_ptr_array_type::tag (), s, snext);
        break;
      case shape_type::Edge:
        erase_shapes_by_tag (typename shape_type::edge_type::tag (), s, snext);
        break;
      case shape_type::Path:
        erase_shapes_by_tag (typename shape_type::path_type::tag (), s, snext);
        break;
      case shape_type::PathRef:
        erase_shapes_by_tag (typename shape_type::path_ref_type::tag (), s, snext);
        break;
      case shape_type::PathPtrArrayMember:
        if (! whole_array) { 
          resolve_array_except (typename shape_type::path_ptr_array_type::tag (), typename shape_type::path_ref_type::tag (), s, snext);
          break;
        }
      case shape_type::PathPtrArray:
        erase_shapes_by_tag (typename shape_type::path_ptr_array_type::tag (), s, snext);
        break;
      case shape_type::Box:
        erase_shapes_by_tag (typename shape_type::box_type::tag (), s, snext);
        break;
      case shape_type::BoxArrayMember:
        if (! whole_array) { 
          resolve_box_array_except (typename shape_type::box_array_type::tag (), typename shape_type::box_type::tag (), s, snext);
          break;
        }
      case shape_type::BoxArray:
        erase_shapes_by_tag (typename shape_type::box_array_type::tag (), s, snext);
        break;
      case shape_type::ShortBox:
        erase_shapes_by_tag (typename shape_type::short_box_type::tag (), s, snext);
        break;
      case shape_type::ShortBoxArrayMember:
        if (! whole_array) { 
          resolve_box_array_except (typename shape_type::short_box_array_type::tag (), typename shape_type::short_box_type::tag (), s, snext);
          break;
        }
      case shape_type::ShortBoxArray:
        erase_shapes_by_tag (typename shape_type::short_box_array_type::tag (), s, snext);
        break;
      case shape_type::Text:
        erase_shapes_by_tag (typename shape_type::text_type::tag (), s, snext);
        break;
      case shape_type::TextRef:
        erase_shapes_by_tag (typename shape_type::text_ref_type::tag (), s, snext);
        break;
      case shape_type::TextPtrArrayMember:
        if (! whole_array) { 
          resolve_array_except (typename shape_type::text_ptr_array_type::tag (), typename shape_type::text_ref_type::tag (), s, snext);
          break;
        }
      case shape_type::TextPtrArray:
        erase_shapes_by_tag (typename shape_type::text_ptr_array_type::tag (), s, snext);
        break;
      case shape_type::UserObject:
        erase_shapes_by_tag (typename shape_type::user_object_type::tag (), s, snext);
        break;
      };

      s = snext;

    }
  }

  /**
   *  @brief Erase an element 
   *
   *  Erases a shapes at the given positions [from,to)
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param from The first position of the shapes to erase
   *  @param to The past-end position of the shapes to erase
   */
  template <class Tag> 
  void erase (Tag tag, typename db::layer<typename Tag::shape_type>::iterator from,
                       typename db::layer<typename Tag::shape_type>::iterator to)
  {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, from, to));
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    layer<typename Tag::shape_type> ().erase (from, to);
  }

  /**
   *  @brief Erasing of multiple elements
   *
   *  Erase a set of positions given by an iterator I: *(from,to).
   *  *I must render an "iterator" object.
   *  The iterators in the sequence from, to must be sorted in
   *  "later" order.
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param first The start of the sequence of iterators
   *  @param last The end of the sequence of iterators
   */
  template <class Tag, class I>
  void erase_positions (Tag tag, I first, I last)
  {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, first, last, true /*dummy*/));
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    layer<typename Tag::shape_type> ().erase_positions (first, last);
  }

  /**
   *  @brief Replace an element at the given position 
   *
   *  Replaces the element at the position pos with the
   *  new element.
   *  Replacing a shape will invalidate the bbox and the sorting
   *  state.
   *
   *  @param pos The position at which to replace the shape
   *  @param sh The shape to replace 
   *  
   *  @return A reference to the object created
   */
  template <class Sh>
  const Sh &replace (typename db::layer<Sh>::iterator pos, const Sh &sh)
  {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, Sh> (false /*not insert*/, *pos));
      manager ()->queue (this, new db::layer_op<shapes, Sh> (true /*insert*/, sh));
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    return layer<Sh> ().replace (pos, sh);
  }

  /**
   *  @brief updates the bbox 
   *
   *  Updating the bbox is required after insert operations
   *  and is performed only as far as necessary.
   */
  void update_bbox ()
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->update_bbox ();
    }
  }

  /**
   *  @brief check if the bounding box needs update
   *
   *  Returns true if the bounding box of the shapes has changed and 
   *  requires an update.
   */
  bool is_bbox_dirty () const
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      if ((*l)->is_bbox_dirty ()) {
        return true;
      }
    }
    return false;
  }

  /**
   *  @brief Retrieve the bbox 
   *
   *  Retrieving the bbox might required an update_bbox
   *  before the bbox is valid. It will assert if the bbox
   *  was not valid.
   *  Doing a update_bbox before does not imply a performance penalty
   *  since the state is cached.
   */
  box_type bbox () const
  {
    box_type box;
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      box += (*l)->bbox ();
    }
    return box;
  }

  /**
   *  @brief sorts the trees
   *
   *  Sorting the trees is required after insert operations
   *  and is performed only as far as necessary.
   */
  void sort () 
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->sort ();
    }
  }

  /**
   *  @brief Clears the collection
   */
  void clear ()
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->clear (this, manager ());
      delete *l;
    }
    invalidate_state ();  //  HINT: must come before the change is done!
    m_layers.clear ();
  }

  /**
   *  @brief Do a region search for a certain shape type in "touching" mode
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The region iterator 
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::touching_iterator begin_touching (const box_type &b, Tag type) const 
  {
    return layer<typename Tag::shape_type> ().begin_touching (b);
  }

  /**
   *  @brief The region iterator end token
   */
  template <class Tag> 
  size_t end_touching (Tag type) const 
  {
    return 0; // KLUDGE: some shortcut but basically an implementation detail ..
  }

  /**
   *  @brief Do a region search for a certain shape type in "overlapping" mode
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The region iterator 
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::overlapping_iterator begin_overlapping (const box_type &b, Tag type) const 
  {
    return layer<typename Tag::shape_type> ().begin_overlapping (b);
  }

  /**
   *  @brief The region iterator end token
   */
  template <class Tag> 
  size_t end_overlapping (Tag type) const 
  {
    return 0; // KLUDGE: some shortcut but basically an implementation detail ..
  }

  /**
   *  @brief begin iterator of all elements of a certain shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The first position of the shapes of the requested type
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::iterator begin (Tag type) const
  {
    return layer<typename Tag::shape_type> ().begin ();
  }

  /**
   *  @brief end iterator of all elements of a certain shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The post-end position of the shapes of the requested type
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::iterator end (Tag type) const
  {
    return layer<typename Tag::shape_type> ().end ();
  }

  /**
   *  @brief begin iterator of all elements 
   *
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin (unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief begin iterator of all elements with a touching mode region query
   *
   *  @param box The region to query the shapes from
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin_touching (const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, box, shape_iterator::Touching, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief begin iterator of all elements with a overlapping mode region query
   *
   *  @param box The region to query the shapes from
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin_overlapping (const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, box, shape_iterator::Overlapping, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief find a given shape (exactly)
   *
   *  @param s The shape to find
   *  
   *  @return end(Sh::tag) if the shape was not found, the position otherwise
   */
  template <class Sh> 
  typename db::layer<Sh>::iterator find (const Sh &s) const
  {
    return layer<Sh> ().find (s);
  }

  /** 
   *  @brief Implementation of the redo method
   */
  void redo (db::Op *op) throw ()
  {
    db::layer_op_base<shapes> *layop = dynamic_cast< db::layer_op_base<shapes> *> (op);
    if (layop) {
      layop->redo (this);
    } 
  }

  /** 
   *  @brief Implementation of the undo method
   */
  void undo (db::Op *op) throw ()
  {
    db::layer_op_base<shapes> *layop = dynamic_cast< db::layer_op_base<shapes> *> (op);
    if (layop) {
      layop->undo (this);
    } 
  }

  void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.cell_info (m_layers);
    if (mp_state_model) {
      m.cell_info (mp_state_model);
    }

    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->collect_mem_stat (m);
    }
  }

private:
  template <class Sh>
  const db::layer<Sh> &layer () const
  {
    typedef layer_class<shapes, Sh> lay_cls;
    
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      const lay_cls *lc = dynamic_cast <const lay_cls *> (*l);
      if (lc) {
        return lc->layer ();
      }
    }

    //  there seems to be a bug in gcc that disables
    //  the use of a simple static object here:
    static const db::layer<Sh> *empty_layer = 0;
    if (! empty_layer) {
      empty_layer = new db::layer<Sh> ();
    }
    return *empty_layer;
  }

  template <class Sh>
  db::layer<Sh> &layer () 
  {
    typedef layer_class<shapes, Sh> lay_cls;
    lay_cls *lc;
    
    for (typename tl::vector<layer_base<shapes> *>::iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      lc = dynamic_cast <lay_cls *> (*l);
      if (lc) {
        //  this is what optimizes access times for another access
        //  with this type 
        std::swap (m_layers.front (), *l);
        return lc->layer ();
      }
    }

    //  create 
    lc = new lay_cls ();
    m_layers.push_back (lc);
    std::swap (m_layers.front (), m_layers.back ());
    return lc->layer ();
  }

  void invalidate_state ()
  {
    if (mp_state_model) {
      mp_state_model->invalidate_bboxes ();
    }
  }

  /**
   *  @brief (Internal) Resolve an array and retain a single instance
   */
  template <class STag, class ATag> 
  void resolve_array_except (ATag atag, STag stag, const shape_type &shape)
  {
    typedef typename ATag::shape_type array_type;
    typedef typename STag::shape_type single_ref_type;

    bool with_props = shape.has_prop_id ();

    //  Resolve the array without using the specified element
    array_type *array = with_props ? ((db::shape_with_properties<array_type> *) shape.basic_ptr ()) : ((array_type *) shape.basic_ptr ());

    for (typename array_type::iterator ainst = array->begin (); !ainst.at_end (); ++ainst) {
      //  Hint: this assumes that single_ref_type is a reference with a displacement only and
      //  array_type uses a simple transformation without rotation.
      if ((*ainst).disp () != shape.array_trans ().disp ()) {
        //  An instance to keep
        if (with_props) {
          this->insert (db::shape_with_properties<single_ref_type> (single_ref_type (&array->object ().obj (), (*ainst).disp ()), shape.prop_id ()));
        } else {
          this->insert (single_ref_type (&array->object ().obj (), (*ainst).disp ()));
        }
      }
    }

    //  erase the array object itself
    erase_shape_by_tag (atag, shape);
  }

  /**
   *  @brief (Internal) Resolve an box array and retain a single instance
   */
  template <class STag, class ATag>
  void resolve_box_array_except (ATag atag, STag stag, const shape_type &shape)
  {
    typedef typename ATag::shape_type array_type;
    typedef typename STag::shape_type single_box_type;

    bool with_props = shape.has_prop_id ();

    //  Resolve the array without using the specified element
    array_type *array = with_props ? ((db::shape_with_properties<array_type> *) shape.basic_ptr ()) : ((array_type *) shape.basic_ptr ());

    for (typename array_type::iterator ainst = array->begin (); !ainst.at_end (); ++ainst) {
      //  Hint: this assumes that array_type uses a disp_trans only
      if ((*ainst).disp () != shape.array_trans ().disp ()) {
        //  An instance to keep
        if (with_props) {
          this->insert (db::shape_with_properties<single_box_type> (array->object ().transformed (*ainst), shape.prop_id ()));
        } else {
          //  Hint: this form does not preserve the boxes representation, i.e. short. 
          //  However, this may be desired since short box arrays may be used to represent boxes outside
          //  16 bit coordinate space.
          this->insert (array->object ().transformed (*ainst));
        }
      }
    }

    //  erase the array object itself
    erase_shape_by_tag (atag, shape);
  }

  /** 
   *  @brief (Internal) Erase by generic pointer
   */
  template <class Tag>
  void erase_shape_by_tag (Tag tag, const shape_type &shape)
  {
    if (! shape.has_prop_id ()) {
      typename db::layer<typename Tag::shape_type>::iterator b = layer<typename Tag::shape_type> ().begin ();
      erase (Tag (), b + ((typename Tag::shape_type *) shape.basic_ptr () - &*b));
    } else {
      typedef db::shape_with_properties<typename Tag::shape_type> swp_type;
      typename db::layer<swp_type>::iterator b = layer<swp_type> ().begin ();
      erase (typename swp_type::tag (), b + ((swp_type *) shape.basic_ptr () - &*b));
    }
  }

  /** 
   *  @brief (Internal) Erase by generic pointer set
   */
  template <class Tag>
  void erase_shapes_by_tag (Tag tag, typename std::vector<shape_type>::const_iterator s1, typename std::vector<shape_type>::const_iterator s2)
  {
    if (! s1->has_prop_id ()) {

      typename db::layer<typename Tag::shape_type>::iterator b = layer<typename Tag::shape_type> ().begin ();
      std::vector<typename db::layer<typename Tag::shape_type>::iterator> iters;
      iters.reserve (std::distance (s1, s2));

      for (typename std::vector<shape_type>::const_iterator s = s1; s != s2; ++s) {
        typename db::layer<typename Tag::shape_type>::iterator i = b + ((typename Tag::shape_type *) s->basic_ptr () - &*b);
        //  in the "whole array" case it may happen that we delete one object multiple times ..
        if (iters.empty () || ! (iters.back () == i)) {
          iters.push_back (i);
        }
      }

      erase_positions (Tag (), iters.begin (), iters.end ());

    } else {

      typedef db::shape_with_properties<typename Tag::shape_type> swp_type;
      typename db::layer<swp_type>::iterator b = layer<swp_type> ().begin ();
      std::vector<typename db::layer<swp_type>::iterator> iters;
      iters.reserve (std::distance (s1, s2));

      for (typename std::vector<shape_type>::const_iterator s = s1; s != s2; ++s) {
        typename db::layer<swp_type>::iterator i = b + ((swp_type *) s->basic_ptr () - &*b);
        //  in the "whole array" case it may happen that we delete one object multiple times ..
        if (iters.empty () || ! (iters.back () == i)) {
          iters.push_back (i);
        }
      }

      erase_positions (typename swp_type::tag (), iters.begin (), iters.end ());

    }
  }

  /**
   *  @brief (Internal) Resolve an array and retain a single instance
   */
  template <class STag, class ATag> 
  void resolve_array_except (ATag atag, STag stag, typename std::vector<shape_type>::const_iterator s1, typename std::vector<shape_type>::const_iterator s2)
  {
    typedef typename ATag::shape_type array_type;
    typedef typename STag::shape_type single_ref_type;

    if (s1 == s2) {
      return;
    }

    bool with_props = s1->has_prop_id ();

    typename std::vector<shape_type>::const_iterator snext;
    for (typename std::vector<shape_type>::const_iterator s = s1; s != s2; s = snext) {

      //  Resolve the array without using the specified element
      array_type *array = with_props ? ((db::shape_with_properties<array_type> *) s->basic_ptr ()) : ((array_type *) s->basic_ptr ());
      snext = s;
      while (++snext != s2) {
        if (array != (with_props ? ((db::shape_with_properties<array_type> *) snext->basic_ptr ()) : ((array_type *) snext->basic_ptr ()))) {
          break;
        }
      }

      for (typename array_type::iterator ainst = array->begin (); !ainst.at_end (); ++ainst) {
        //  Hint: this assumes that single_ref_type is a reference with a displacement only and
        //  array_type uses a simple transformation without rotation.
        if (! std::binary_search (s, snext, (*ainst).disp (), compare_array_displacement<shape_type> ())) {
          //  An instance to keep
          if (with_props) {
            this->insert (db::shape_with_properties<single_ref_type> (single_ref_type (&array->object ().obj (), (*ainst).disp ()), s->prop_id ()));
          } else {
            this->insert (single_ref_type (&array->object ().obj (), (*ainst).disp ()));
          }
        }
      }

    }

    //  erase the array object itself
    erase_shapes_by_tag (atag, s1, s2);
  }

  /**
   *  @brief (Internal) Resolve an box array and retain a single instance
   */
  template <class STag, class ATag>
  void resolve_box_array_except (ATag atag, STag stag, typename std::vector<shape_type>::const_iterator s1, typename std::vector<shape_type>::const_iterator s2)
  {
    typedef typename ATag::shape_type array_type;
    typedef typename STag::shape_type single_box_type;

    if (s1 == s2) {
      return;
    }

    bool with_props = s1->has_prop_id ();

    typename std::vector<shape_type>::const_iterator snext;
    for (typename std::vector<shape_type>::const_iterator s = s1; s != s2; s = snext) {

      //  Resolve the array without using the specified element
      array_type *array = with_props ? ((db::shape_with_properties<array_type> *) s->basic_ptr ()) : ((array_type *) s->basic_ptr ());
      snext = s;
      while (++snext != s2) {
        if (array != (with_props ? ((db::shape_with_properties<array_type> *) snext->basic_ptr ()) : ((array_type *) snext->basic_ptr ()))) {
          break;
        }
      }

      for (typename array_type::iterator ainst = array->begin (); !ainst.at_end (); ++ainst) {
        //  Hint: since we have sorted Shape references, we can expect that the transformations are
        //  sorted and we can look for a certain transformation by binary search:
        if (! std::binary_search (s, snext, (*ainst).disp (), compare_array_displacement<shape_type> ())) {
          //  An instance to keep
          if (with_props) {
            this->insert (db::shape_with_properties<single_box_type> (array->object ().transformed (*ainst), s->prop_id ()));
          } else {
            //  Hint: this form does not preserve the boxes representation, i.e. short. 
            //  However, this may be desired since short box arrays may be used to represent boxes outside
            //  16 bit coordinate space.
            this->insert (array->object ().transformed (*ainst));
          }
        }
      }

    }

    //  erase the array object itself
    erase_shapes_by_tag (atag, s1, s2);
  }

  tl::vector<layer_base<shapes> *> m_layers;
  db::LayoutStateModel *mp_state_model;
};

/**
 *  @brief A undo/redo queue object for the layer
 *
 *  This class is used internally to queue an insert or erase operation
 *  into the db::Object manager's undo/redo queue.
 */
template <class Shapes, class Sh>
class layer_op
  : public layer_op_base<Shapes> 
{
public:
  layer_op (bool insert, const Sh &sh)
    : m_insert (insert)
  {
    m_shapes.push_back (sh);
  }
  
  template <class Iter>
  layer_op (bool insert, Iter from, Iter to)
    : m_insert (insert)
  {
    m_shapes.insert (m_shapes.end (), from, to);
  }

  template <class Iter>
  layer_op (bool insert, Iter from, Iter to, bool dummy)
    : m_insert (insert)
  {
    m_shapes.reserve (std::distance (from, to));
    for (Iter i = from; i != to; ++i) {
      m_shapes.push_back (**i);
    }
  }

  virtual void undo (Shapes *shapes)
  {
    if (m_insert) {
      erase (shapes);
    } else {
      insert (shapes);
    }
  }

  virtual void redo (Shapes *shapes)
  {
    if (m_insert) {
      insert (shapes);
    } else {
      erase (shapes);
    }
  }

private:
  bool m_insert;
  std::vector<Sh> m_shapes;

  void insert (Shapes *shapes)
  {
    shapes->insert (m_shapes.begin (), m_shapes.end ());
  }

  void erase (Shapes *shapes)
  {
    if (size_t (std::distance (shapes->begin (typename Sh::tag ()), shapes->end (typename Sh::tag ()))) <= m_shapes.size ()) {
      //  If all shapes are to be removed, just clear the shapes
      shapes->erase (typename Sh::tag (), shapes->begin (typename Sh::tag ()), shapes->end (typename Sh::tag ()));
    } else {
      std::sort (m_shapes.begin (), m_shapes.end ());
      //  This is not quite effective but seems to be the simpliest way
      //  of implementing this: search for each element and erase these.
      std::vector<typename db::layer<Sh>::iterator> to_erase;
      to_erase.reserve (m_shapes.size ());
      for (typename db::layer<Sh>::iterator lsh = shapes->begin (typename Sh::tag ()); lsh != shapes->end (typename Sh::tag ()); ++lsh) {
        typename std::vector<Sh>::const_iterator s = std::lower_bound (m_shapes.begin (), m_shapes.end (), *lsh);
        if (s != m_shapes.end () && *s == *lsh) {
          to_erase.push_back (lsh);
        }
      }
      shapes->erase_positions (typename Sh::tag (), to_erase.begin (), to_erase.end ());
    }
  }
};

/** 
 *  @brief A typedef for a standard shapes container 
 */
typedef db::shapes<db::Coord> Shapes;

/** 
 *  @brief A typedef for the floating-point coordinate shapes container
 */
typedef db::shapes<db::DCoord> DShapes;

}  // namespace db
  
#endif

