
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtInstPropertiesPage.h"
#include "edtUtils.h"
#include "layLayoutView.h"
#include "tlString.h"

#include <string>

namespace edt
{


// -------------------------------------------------------------------------
//  helper function to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

// -------------------------------------------------------------------------
//  InstPropertiesPage implementation

InstPropertiesPage::InstPropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), mp_service (service), m_enable_cb_callback (true)
{
  m_pos = service->selection ().begin ();
  m_index = 0;
  mp_service->clear_highlights ();

  setupUi (this);
  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (sel_pb, SIGNAL (clicked ()), this, SLOT (show_cell ()));
  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));

  for (unsigned int r = 0; r < 8; ++r) {
    orient_cmb->addItem (db::fixpoint_trans (r).to_string ().c_str ());
  }
}

InstPropertiesPage::~InstPropertiesPage ()
{
  mp_service->restore_highlights ();
}

void
InstPropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  mp_service->view ()->dbu_coordinates (dbu_cb->isChecked ());
  mp_service->view ()->absolute_coordinates (abs_cb->isChecked ());

  update ();
}

void 
InstPropertiesPage::back ()
{
  m_pos = mp_service->selection ().end ();
  m_index = mp_service->selection ().size ();
}

void 
InstPropertiesPage::front ()
{
  m_pos = mp_service->selection ().begin ();
  m_index = 0;
}

bool 
InstPropertiesPage::at_begin () const
{
  return (m_pos == mp_service->selection ().begin ());
}

bool 
InstPropertiesPage::at_end () const
{
  return (m_pos == mp_service->selection ().end ());
}

void 
InstPropertiesPage::operator-- ()
{
  --m_pos;
  --m_index;
}

void 
InstPropertiesPage::operator++ ()
{
  ++m_pos;
  ++m_index;
}

void 
InstPropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

void 
InstPropertiesPage::update ()
{
  tl_assert (m_pos->is_cell_inst ());

  mp_service->highlight (m_index);

  m_enable_cb_callback = false;
  dbu_cb->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_cb->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;

  const lay::CellView &cv = mp_service->view ()->cellview (m_pos->cv_index ());
  double dbu = cv->layout ().dbu ();

  std::string info (tl::translate ("Cell "));
  info += cv->layout ().cell_name (m_pos->cell_index ());
  info_lbl->setText (info.c_str ());

  cell_name_le->setText (cv->layout ().cell_name (m_pos->back ().cell_inst->object ().cell_index ()));

  db::Point rowv, columnv;
  unsigned long rows, columns;

  db::CplxTrans gt;
  if (abs_cb->isChecked ()) {
    gt = m_pos->trans () * m_pos->back ().complex_trans ().inverted ();
  }

  bool du = dbu_cb->isChecked ();

  if (m_pos->back ().cell_inst->is_regular_array (rowv, columnv, rows, columns)) {

    array_grp->setChecked (true);
    rows_le->setText (tl::to_string (rows).c_str ());
    columns_le->setText (tl::to_string (columns).c_str ());
    row_x_le->setText (coord_to_string ((gt * rowv).x (), dbu, du).c_str ());
    row_y_le->setText (coord_to_string ((gt * rowv).y (), dbu, du).c_str ());
    column_x_le->setText (coord_to_string ((gt * columnv).x (), dbu, du).c_str ());
    column_y_le->setText (coord_to_string ((gt * columnv).y (), dbu, du).c_str ());

    db::Trans t (m_pos->back ().cell_inst->global_trans ());
    pos_x_le->setText (coord_to_string ((gt * t.disp ()).x (), dbu, du).c_str ());
    pos_y_le->setText (coord_to_string ((gt * t.disp ()).y (), dbu, du).c_str ());
    orient_cmb->setCurrentIndex (t.rot ());
    //  @@@ TODO: implement arbitrary rotation
    mag_le->setText (tl::to_string (m_pos->back ().cell_inst->complex_trans ().mag ()).c_str ());

    //  from the actual position vs. base position try to compute the array indices
    db::Trans t2 (*m_pos->back ().array_inst);

    unsigned long row, column;
    if (rowv.x () != 0) {
      row = (t2.disp ().x () - t.disp ().x ()) / rowv.x (); 
    } else if (rowv.y () != 0) {
      row = (t2.disp ().y () - t.disp ().y ()) / rowv.y (); 
    } else {
      row = 0;
    }
    if (columnv.x () != 0) {
      column = (t2.disp ().x () - t.disp ().x ()) / columnv.x (); 
    } else if (columnv.y () != 0) {
      column = (t2.disp ().y () - t.disp ().y ()) / columnv.y (); 
    } else {
      column = 0;
    }
    inst_lbl->setText (tl::sprintf (tl::translate ("This is instance [%d,%d] of array with").c_str (), row, column).c_str ());

  } else {

    array_grp->setChecked (false);
    rows_le->setText ("");
    columns_le->setText ("");
    row_x_le->setText ("");
    row_y_le->setText ("");
    column_x_le->setText ("");
    column_y_le->setText ("");

    db::Trans t (*m_pos->back ().array_inst);
    pos_x_le->setText (coord_to_string ((gt * t.disp ()).x (), dbu, du).c_str ());
    pos_y_le->setText (coord_to_string ((gt * t.disp ()).y (), dbu, du).c_str ());
    orient_cmb->setCurrentIndex (t.rot ());
    //  @@@ TODO: implement arbitrary rotation
    mag_le->setText (tl::to_string (m_pos->back ().cell_inst->complex_trans ().mag ()).c_str ());
    inst_lbl->setText ("");

  }

}

void 
InstPropertiesPage::show_cell ()
{
  lay::CellView::unspecific_cell_path_type path (mp_service->view ()->cellview (m_pos->cv_index ()).combined_unspecific_path ());
  for (edt::ObjectInstPath::iterator p = m_pos->begin (); p != m_pos->end (); ++p) {
    path.push_back (p->cell_inst->object ().cell_index ());
  }

  mp_service->view ()->set_current_cell_path (m_pos->cv_index (), path);
}

void
InstPropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), *m_pos);
}

bool 
InstPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
InstPropertiesPage::apply ()
{

  //  @@@

}

}

