
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtService.h"
#include "edtPropertiesPages.h"
#include "edtInstPropertiesPage.h"
#include "layLayoutView.h"
#include "tlProgress.h"

namespace edt
{

// -------------------------------------------------------------

//  search range for select in pixels
unsigned int search_range = 5;   // @@@ make variable?

//  search range for select in pixels
unsigned int inst_search_range = 5;   // @@@ make variable?

//  max. number of tries in single-click selection before giving up
int point_sel_tests = 10000;

//  max. number of tries in single-click selection before giving up
int inst_point_sel_tests = 10000;

// -------------------------------------------------------------

class ShapeFinder
  : public Finder
{
public:
  struct StopException { };

  typedef std::vector<edt::ObjectInstPath> founds_vector_type;
  typedef founds_vector_type::const_iterator iterator;

  ShapeFinder (bool point_mode, db::ShapeIterator::flags_type flags)
    : Finder (point_mode), m_flags (flags), mp_prop_sel (0), m_inv_prop_sel (false), m_progress (tl::translate ("Selecting ..."))
  {
    m_progress.set_unit (1000);
    m_progress.set_format ("");
  }

  bool find (lay::LayoutView *view, const lay::LayerProperties &lprops, const db::DBox &region_mu)
  {
    m_cv_index = lprops.cellview_index ();
    const lay::CellView &cv = view->cellview (m_cv_index);
    if (! cv.is_valid ()) {
      return false;
    }

    m_topcell = cv.cell_index ();

    double dbu = cv->layout ().dbu ();
    db::Box region = db::Box::from_double (cv.context_trans ().inverted () * (region_mu * (1.0 / dbu)));

    m_tries = point_sel_tests;
    mp_prop_sel = &lprops.prop_sel ();
    m_inv_prop_sel = lprops.inverse_prop_sel ();

    //  actually find
    try {
      start (view, m_cv_index, lprops.trans (), cv->layout (), region, *cv.cell (), view->get_min_hier_levels (), view->get_max_hier_levels (), int (lprops.layer_index ()));
    } catch (StopException) {
      return false;
    }

    //  return true if anything was found
    return ! m_founds.empty ();
  }

  iterator begin () const
  {
    return m_founds.begin ();
  }

  iterator end () const
  {
    return m_founds.end ();
  }

private:
  virtual void 
  visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int /*level*/)
  {
    if (! point_mode ()) {

      ++m_progress;

      const db::Shapes &shapes = cell.shapes (layer ());

      db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
      while (! shape.at_end ()) {

        ++m_progress;

        //  in box mode, just test the boxes
        if (shape->bbox ().inside (search_box)) {

          //  in point mode just store that found that has the least "distance"
          m_founds.push_back (edt::ObjectInstPath ());
          m_founds.back ().set_cv_index (m_cv_index);
          m_founds.back ().set_topcell (m_topcell);
          m_founds.back ().assign_path (path ().begin (), path ().end ());
          m_founds.back ().set_layer (layer ());
          m_founds.back ().set_shape (*shape);

        }

        ++shape;

      } 

    } else {

      if (--m_tries < 0) {
        throw StopException ();
      }

      const db::Shapes &shapes = cell.shapes (layer ());

      db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
      while (! shape.at_end ()) {

        bool match = false;
        double d = std::numeric_limits<double>::max ();

        if (--m_tries < 0) {
          throw StopException ();
        }

        db::Point point (search_box.center ());

        //  in point mode, test the edges and use a "closest" criterion
        if (shape->is_polygon ()) {

          for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e, d, match);
          }

          //  test if inside the polygon
          if (! match && inside_poly (shape->begin_edge (), point) >= 0) { 
            d = t.ctrans (poly_dist (shape->begin_edge (), point)); 
            match = true;
          }

        } else if (shape->is_path ()) {

          //  test the "spine"
          db::Shape::point_iterator pt = shape->begin_point (); 
          if (pt != shape->end_point ()) {
            db::Point p (*pt);
            ++pt;
            for (; pt != shape->end_point (); ++pt) {
              test_edge (t * db::Edge (p, *pt), d, match);
              p = *pt;
            }
          }

          //  convert to polygon and test those edges
          db::Polygon poly;
          shape->polygon (poly);
          for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e, d, match);
          }

          //  test if inside the polygon
          if (! match && inside_poly (poly.begin_edge (), point) >= 0) { 
            d = t.ctrans (poly_dist (poly.begin_edge (), point));
            match = true;
          }

        } else if (shape->is_box ()) {

          const db::Box &box = shape->box ();

          //  convert to polygon and test those edges
          db::Polygon poly (box);
          for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e, d, match);
          }

          if (! match && box.contains (search_box.center ())) {
            d = t.ctrans (poly_dist (poly.begin_edge (), point)); 
            match = true;
          }

        } else if (shape->is_text ()) {

          db::Point tp (shape->text_trans () * db::Point ());
          if (search_box.contains (tp)) {
            d = tp.distance (search_box.center ());
            match = true;
          }

        }

        if (match && closer (d)) {

          //  in point mode just store that found that has the least "distance"
          if (m_founds.empty ()) {
            m_founds.push_back (edt::ObjectInstPath ());
          }

          m_founds.back ().set_cv_index (m_cv_index);
          m_founds.back ().set_topcell (m_topcell);
          m_founds.back ().assign_path (path ().begin (), path ().end ());
          m_founds.back ().set_layer (layer ());
          m_founds.back ().set_shape (*shape);

        }

        ++shape;

      }

    }

  }

  std::vector<edt::ObjectInstPath> m_founds;
  db::ShapeIterator::flags_type m_flags;
  unsigned int m_cv_index;
  db::Layout::cell_index_type m_topcell;
  const std::set<db::PropertiesRepository::properties_id_type> *mp_prop_sel;
  bool m_inv_prop_sel;
  int m_tries;
  tl::Progress m_progress;
};

// -------------------------------------------------------------

class InstFinder
  : public edt::Finder
{
public:
  struct StopException { };

  typedef std::vector<edt::ObjectInstPath> founds_vector_type;
  typedef founds_vector_type::const_iterator iterator;

  InstFinder (bool point_mode)
    : Finder (point_mode), m_progress (tl::translate ("Selecting ..."))
  {
    m_progress.set_unit (1000);
    m_progress.set_format ("");
  }

  bool find (lay::LayoutView *view, unsigned int cv_index, const db::CplxTrans &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level)
  {
    m_cv_index = cv_index;
    m_topcell = cell.cell_index ();
    m_tries = inst_point_sel_tests;
    mp_view = view;

    //  actually find
    try {
      std::vector<db::CplxTrans> tv;
      tv.push_back (trans);
      start (view, cv_index, tv, layout, region, cell, min_level, max_level);
    } catch (StopException) {
      return false;
    }

    //  return true if anything was found
    return ! m_founds.empty ();
  }
 
  iterator begin () const
  {
    return m_founds.begin ();
  }

  iterator end () const
  {
    return m_founds.end ();
  }

private:
  virtual void 
  visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int level)
  {
    if (! point_mode ()) {

      ++m_progress;

      //  shortcut for cases where nothing can be found
      if (level != max_level () - 1 && (mp_view->hidden_cells ().size () <= m_cv_index || mp_view->hidden_cells () [m_cv_index].empty ())) {
        return;
      }

      //  look for instances to check here ..
      db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = *inst;

        ++m_progress;

        //  just consider the instances exactly at the last level of 
        //  hierarchy (this is where the boxes are drawn) or if of cells that
        //  are hidden.
        if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

          db::box_convert <db::CellInst> bc (layout ());
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
          
            ++m_progress;

            const db::DBox box = cell_inst.complex_trans (*p) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

            //  in box mode, just test the boxes
            if (box.inside (db::DBox (search_box))) {

              //  in point mode just store that found that has the least "distance"
              m_founds.push_back (edt::ObjectInstPath ());
              m_founds.back ().set_cv_index (m_cv_index);
              m_founds.back ().set_topcell (m_topcell);
              m_founds.back ().assign_path (path ().begin (), path ().end ());

              //  add the selected instance as the last element of the path
              db::InstElement el;
              el.cell_inst = &cell_inst;
              el.array_inst = p;
              m_founds.back ().add_path (el);

            }

          }

        }

        ++inst;

      }

    } else {

      if (--m_tries < 0) {
        throw StopException ();
      }

      //  look for instances to check here ..
      db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
      while (! inst.at_end ()) {

        if (--m_tries < 0) {
          throw StopException ();
        }

        const db::CellInstArray &cell_inst = *inst;

        //  just consider the instances exactly at the last level of 
        //  hierarchy (this is where the boxes are drawn) or if of cells that
        //  are hidden.
        if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

          db::box_convert <db::CellInst> bc (layout ());
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
          
            if (--m_tries < 0) {
              throw StopException ();
            }

            bool match = false;
            double d = std::numeric_limits<double>::max ();

            const db::DBox box = cell_inst.complex_trans (*p) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

            //  convert to polygon and test those edges
            db::DPolygon poly (box);
            for (db::DPolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
              test_edge (t * *e, d, match);
            }

            if (! match && box.contains (search_box.center ())) {
              d = t.ctrans (poly_dist (poly.begin_edge (), db::DPoint (search_box.center ()))); 
              match = true;
            }

            if (match && closer (d)) {

              //  in point mode just store that found that has the least "distance"
              if (m_founds.empty ()) {
                m_founds.push_back (edt::ObjectInstPath ());
              }

              m_founds.back ().set_cv_index (m_cv_index);
              m_founds.back ().set_topcell (m_topcell);
              m_founds.back ().assign_path (path ().begin (), path ().end ());

              //  add the selected instance as the last element of the path
              db::InstElement el;
              el.cell_inst = &cell_inst;
              el.array_inst = p;
              m_founds.back ().add_path (el);

            }

          }

        }

        ++inst;

      }

    }

  }

  unsigned int m_cv_index;
  db::Layout::cell_index_type m_topcell;
  std::vector<edt::ObjectInstPath> m_founds;
  int m_tries;
  tl::Progress m_progress;
  lay::LayoutView *mp_view;
};


// -------------------------------------------------------------

Service::Service (db::Manager *manager, lay::LayoutView *view, db::ShapeIterator::flags_type flags)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    m_min_level (0), m_max_level (0),
    m_cell_inst_service (false),
    m_flags (flags),
    m_line_width (1), m_vertex_size (3)
{ 
  //  .. nothing yet ..
}

Service::Service (db::Manager *manager, lay::LayoutView *view)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    m_min_level (0), m_max_level (0),
    m_cell_inst_service (true),
    m_flags (db::ShapeIterator::Nothing),
    m_line_width (1), m_vertex_size (3)
{ 
  //  .. nothing yet ..
}

Service::~Service ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();
}

void 
Service::clear_highlights ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (false);
  }
}

void 
Service::restore_highlights ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (true);
  }
}

void 
Service::highlight (unsigned int n)
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (n-- == 0);
  }
}

bool 
Service::has_selection ()
{
  return ! m_selection.empty ();
}

double 
Service::click_proximity (const db::DPoint &pos)
{
  if (m_cell_inst_service) {

    InstFinder finder (true);

    //  go through all cell views
    std::set< std::pair<db::CplxTrans, int> > variants = mp_view->cv_transform_variants ();
    for (std::set< std::pair<db::CplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {

      const lay::CellView &cv = mp_view->cellview (v->second);
      if (cv.is_valid ()) {

        double dbu = cv->layout ().dbu ();

        //  compute search box
        double l = double (inst_search_range) / (widget ()->mouse_event_trans () * db::DCplxTrans (v->first)).mag ();
        db::Box search_box = db::Box::from_double (db::DBox (pos, pos).enlarged (db::DPoint (l, l)) * (1.0 / dbu));

        finder.find (mp_view, v->second, v->first, cv->layout (), search_box, *cv.cell (), int (m_min_level), int (m_max_level));

      }

    }

    //  Return the finder's proximity value
    if (finder.begin () != finder.end ()) {
      return finder.proximity ();
    } else {
      return lay::Editable::click_proximity (pos); 
    }

  } else {

    ShapeFinder finder (true, m_flags);

    //  go through all visible layers of all cellviews
    for (lay::LayerPropertiesConstIterator lprop = mp_view->begin_layers (); ! lprop.at_end () ; ++lprop) {

      if (lprop->visible (true /*real*/)) {

        //  compute search box
        double l = double (search_range) / (widget ()->mouse_event_trans () * db::DCplxTrans (lprop->trans ()[0])).mag ();
        db::DBox search_box = db::DBox (pos, pos).enlarged (db::DPoint (l, l));

        finder.find (mp_view, *lprop, search_box);

      }

    }

    //  Return the finder's proximity value
    if (finder.begin () != finder.end ()) {
      return finder.proximity ();
    } else {
      return lay::Editable::click_proximity (pos); 
    }

  }
}

bool 
Service::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selection.empty ()) {
      m_selection.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selection.empty ()) {
        m_selection.clear ();
        needs_update = true;
      }
    } else {

      //  extract all shapes

      //  @@@ not implemented yet @@@

    }

  } else if (m_cell_inst_service) {

    InstFinder finder (box.is_point ());

    //  go through all cell views
    std::set< std::pair<db::CplxTrans, int> > variants = mp_view->cv_transform_variants ();
    for (std::set< std::pair<db::CplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {

      const lay::CellView &cv = mp_view->cellview (v->second);
      if (cv.is_valid ()) {

        double dbu = cv->layout ().dbu ();

        //  compute search box
        double l = double (inst_search_range) / (widget ()->mouse_event_trans () * db::DCplxTrans (v->first)).mag ();
        db::Box search_box = db::Box::from_double (box.enlarged (db::DPoint (l, l)) * (1.0 / dbu));

        finder.find (mp_view, v->second, v->first, cv->layout (), search_box, *cv.cell (), int (m_min_level), int (m_max_level));

      }

    }

    //  collect the founds from the finder
    for (InstFinder::iterator f = finder.begin (); f != finder.end (); ++f) {
      select (*f, mode);
      needs_update = true;
      any_selected = true;
    }

  } else {

    ShapeFinder finder (box.is_point (), m_flags);

    //  go through all visible layers of all cellviews
    for (lay::LayerPropertiesConstIterator lprop = mp_view->begin_layers (); ! lprop.at_end (); ++lprop) {

      //  box-selection
      if (lprop->visible (true /*real*/)) {

        //  compute search box
        double l = double (search_range) / (widget ()->mouse_event_trans () * db::DCplxTrans (lprop->trans ()[0])).mag ();
        db::DBox search_box = box.enlarged (db::DPoint (l, l));

        finder.find (mp_view, *lprop, search_box);

      }

    }

    //  collect the founds from the finder
    for (ShapeFinder::iterator f = finder.begin (); f != finder.end (); ++f) {
      select (*f, mode);
      needs_update = true;
      any_selected = true;
    }
      
  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  return any_selected;
}

lay::PropertiesPage *
Service::properties_page (QWidget *parent)
{
  if (m_cell_inst_service) {
    return new edt::InstPropertiesPage (this, parent);
  } else if (m_flags == db::ShapeIterator::Polygons) {
    return new edt::PolygonPropertiesPage (this, parent);
  } else if (m_flags == db::ShapeIterator::Boxes) {
    return new edt::BoxPropertiesPage (this, parent);
  } else if (m_flags == db::ShapeIterator::Paths) {
    return new edt::PathPropertiesPage (this, parent);
  } else if (m_flags == db::ShapeIterator::Texts) {
    return new edt::TextPropertiesPage (this, parent);
  } else {
    return 0;
  }
}

void 
Service::get_selection (std::vector <ObjectInstPath> &sel) const
{
  sel.clear ();
  sel.reserve (m_selection.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::set<ObjectInstPath>::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    sel.push_back (*r);
  }
}

bool 
Service::select (const ObjectInstPath &obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selection.find (obj) == m_selection.end ()) {
      m_selection.insert (obj);
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
    } else {
      m_selection.insert (obj);
    }
    return true;
  }
  return false;
}

void 
Service::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);
}

void
Service::hier_levels (std::pair<unsigned int, unsigned int> l)
{
  m_max_level = l.second;
  m_min_level = l.first;
}

void 
Service::selection_to_view ()
{
  //  the selection objects need to be recreated since we destroyed the old rulers
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();

  //  Hint: this is a lower bound:
  m_markers.reserve (m_selection.size ());

  //  build the transformation variants cache
  //  TODO: move this into LayoutView, where the other transform variant things are?

  std::map <unsigned int, std::vector<db::CplxTrans> > per_cv_tv;
  std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> > per_cv_and_layer_tv;

  for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); !l.at_end (); ++l) {

    if (! l->has_children ()) {

      unsigned int cvi = (l->cellview_index () >= 0) ? (unsigned int) l->cellview_index () : 0;
      std::vector<db::CplxTrans> &tv = per_cv_tv.insert (std::make_pair (cvi, std::vector<db::CplxTrans> ())).first->second;
      tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());

      if (l->layer_index () >= 0) {
        std::vector<db::CplxTrans> &tv = per_cv_and_layer_tv.insert (std::make_pair (std::make_pair (cvi, (unsigned int) l->layer_index ()), std::vector<db::CplxTrans> ())).first->second;
        tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());
      }

    }

  }

  for (std::set<ObjectInstPath>::iterator r = m_selection.begin (); r != m_selection.end (); ++r) {

    const lay::CellView &cv = mp_view->cellview (r->cv_index ());

    if (m_cell_inst_service) {

      std::map <unsigned int, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_tv.find (r->cv_index ());
      if (tv_list != per_cv_tv.end ()) {
      
        for (std::vector<db::CplxTrans>::const_iterator tv = tv_list->second.begin (); tv != tv_list->second.end (); ++tv) {
   
          tl_assert (r->is_cell_inst ());

          lay::Marker *marker = new lay::Marker (mp_view, r->cv_index ());
          db::box_convert<db::CellInst> bc (cv->layout ());
          marker->set (bc (r->back ().cell_inst->object ()), (*tv * cv.context_trans ()) * r->trans ());

          m_markers.push_back (marker);

        }

      }

    } else {

      tl_assert (! r->is_cell_inst ());

      std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_and_layer_tv.find (std::make_pair (r->cv_index (), r->layer ()));
      if (tv_list != per_cv_and_layer_tv.end ()) {

        for (std::vector<db::CplxTrans>::const_iterator tv = tv_list->second.begin (); tv != tv_list->second.end (); ++tv) {
   
          lay::ShapeMarker *marker = new lay::ShapeMarker (mp_view, r->cv_index ());
          marker->set (r->shape (), (*tv * cv.context_trans ()) * r->trans ());

          m_markers.push_back (marker);

        }

      }

    }

  }
}

void 
Service::set_selection (std::vector <ObjectInstPath>::const_iterator s1, std::vector <ObjectInstPath>::const_iterator s2)
{
  m_selection.clear ();
  m_selection.insert (s1, s2);
  selection_to_view ();
}

// -------------------------------------------------------------

} // namespace edt


