
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <dbBox.h>
#include <dbLayout.h>

#include "edtUtils.h"
#include "layCellView.h"
#include "layLayoutView.h"
#include "layMarker.h"
#include "tlException.h"

namespace edt {

// ----------------------------------------------------------------------
//  edt::Finder implementation

Finder::Finder (bool point_mode)
  : m_min_level (0), m_max_level (0),
    mp_layout (0), m_point_mode (point_mode)
{
  m_distance = std::numeric_limits<double>::max ();
}

Finder::~Finder ()
{
  //  .. nothing yet ..
}

bool 
Finder::closer (double d)
{
  if (d <= m_distance) {
    m_distance = d;
    return true;
  } else {
    return false;
  }
}

void 
Finder::start (lay::LayoutView *view, unsigned int cv_index, const std::vector<db::CplxTrans> &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level, int layer)
{
  m_layer = (unsigned int) layer;
  m_region = region; 
  mp_layout = &layout;
  mp_view = view;
  m_cv_index = cv_index;
  m_min_level = min_level;
  m_max_level = max_level;

  if (layer >= 0) {
    m_box_convert = db::box_convert <db::CellInst> (layout, (unsigned int) layer);
    m_cell_box_convert = db::box_convert <db::Cell> ((unsigned int) layer);
  } else {
    m_box_convert = db::box_convert <db::CellInst> (layout);
    m_cell_box_convert = db::box_convert <db::Cell> ();
  }

  m_path.erase (m_path.begin (), m_path.end ());

  for (std::vector<db::CplxTrans>::const_iterator t = trans.begin (); t != trans.end (); ++t) {
    do_find (cell, 0, *t);
  }
}

void
Finder::test_edge (const db::DEdge &edg, double &distance, bool &match)
{
  //  we hit the region with the edge end points - take the closest vertex
  if (m_region.contains (edg.p1 ()) || m_region.contains (edg.p2 ())) {
    double d = std::min (edg.p1 ().distance (m_region.center ()), 
                         edg.p2 ().distance (m_region.center ()));
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
  
  //  if the edge cuts through the active region: test the
  //  edge as a whole
  if (edg.clipped (m_region).first) {
    double d = edg.distance_abs (m_region.center ());
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
}

void
Finder::do_find (const db::Cell &cell, int level, const db::CplxTrans &t)
{
  if (! (level >= m_max_level || ! (t * m_cell_box_convert (cell)).touches (m_region) || mp_view->is_cell_hidden (cell.cell_index (), m_cv_index))) {

    db::Box touch_box (db::Box::from_double (t.inverted () * m_region)); 

    if (level >= m_min_level) {
      visit_cell (cell, touch_box, t, level);
    }

    db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
    while (! inst.at_end ()) {

      const db::CellInstArray &cell_inst = *inst;
      for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, m_box_convert); ! p.at_end (); ++p) {

        m_path.push_back (db::InstElement (cell_inst, p));

        do_find (mp_layout->cell (cell_inst.object ().cell_index ()), 
                 level + 1,
                 t * cell_inst.complex_trans (*p));

        m_path.pop_back ();

      }

      ++inst;

    }

  }
}

// ----------------------------------------------------------------------
//  edt::InstantiationForm implementation

InstantiationForm::InstantiationForm (QWidget *parent)
  : QDialog (parent), mp_view (0), mp_path (0), mp_marker (0), m_enable_cb_callbacks (false)
{
  Ui::InstantiationForm::setupUi (this);

  connect (list, SIGNAL (itemDoubleClicked (QListWidgetItem *)), this, SLOT (double_clicked (QListWidgetItem *)));
  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
}

InstantiationForm::~InstantiationForm ()
{
  if (mp_marker) {
    delete mp_marker;
    mp_marker = 0;
  }
}

void 
InstantiationForm::display_mode_changed (bool)
{
  if (! m_enable_cb_callbacks) {
    return;
  }

  mp_view->dbu_coordinates (dbu_cb->isChecked ());
  mp_view->absolute_coordinates (abs_cb->isChecked ());

  update ();
}

void 
InstantiationForm::double_clicked (QListWidgetItem *item)
{
  int row = list->row (item);
  if (row < 0) {
    return;
  }

  lay::CellView::unspecific_cell_path_type path (mp_view->cellview (mp_path->cv_index ()).combined_unspecific_path ());
  int nrow = 0;
  for (ObjectInstPath::iterator p = mp_path->begin (); p != mp_path->end () && nrow < row; ++p, ++nrow) {
    path.push_back (p->cell_inst->object ().cell_index ());
  }

  mp_view->set_current_cell_path (mp_path->cv_index (), path);

  if (! mp_marker) {
    mp_marker = new lay::Marker (mp_view, mp_path->cv_index ());
  }

  const db::Layout &layout = mp_view->cellview (mp_path->cv_index ())->layout ();
  db::Box box = layout.cell (row == 0 ? mp_path->topcell () : path.back ()).bbox ();

  //  TODO: this does not consider global transformation and variants of this
  db::CplxTrans abs_trans;
  nrow = 0;
  for (ObjectInstPath::iterator p = mp_path->begin (); p != mp_path->end () && nrow < row; ++p, ++nrow) {
    abs_trans = abs_trans * (p->cell_inst->complex_trans (*(p->array_inst)));
  }

  mp_marker->set (box, abs_trans, mp_view->cv_transform_variants (mp_path->cv_index ()));
}

void 
InstantiationForm::show (lay::LayoutView *view, const ObjectInstPath &path)
{
  mp_view = view;
  mp_path = &path;

  m_enable_cb_callbacks = false;
  dbu_cb->setChecked (mp_view->dbu_coordinates ());
  abs_cb->setChecked (mp_view->absolute_coordinates ());
  m_enable_cb_callbacks = true;

  update ();
  exec ();

  mp_view = 0;
  mp_path = 0;
}

void 
InstantiationForm::update ()
{
  bool dbu_coord = dbu_cb->isChecked ();
  bool abs_coord = abs_cb->isChecked ();

  const lay::CellView &cv = mp_view->cellview (mp_path->cv_index ());
  double dbu = cv->layout ().dbu ();

  layout_le->setText (cv->name ().c_str ());

  list->clear ();

  list->addItem (cv->layout ().cell_name (cv.ctx_cell_index ()));
  db::CplxTrans abs_trans;

  //  first include the context path of the cellview in order to tell the path within the cell shown
  for (lay::CellView::specific_cell_path_type::const_iterator p = cv.specific_path ().begin (); p != cv.specific_path ().end (); ++p) {

    //  build the instance information from the path

    db::CplxTrans trans (p->cell_inst->complex_trans (*(p->array_inst)));
    abs_trans = abs_trans * trans;

    if (abs_coord) {
      trans = abs_trans;
    }

    std::string line;
    line += cv->layout ().cell_name (p->cell_inst->object ().cell_index ());
    line += "\tat ";
    line += trans.to_string (true /*lazy*/, dbu_coord ? 0.0 : dbu);
    
    list->addItem (line.c_str ());

  }

  //  then, add the actual path to the object within the target cell
  for (ObjectInstPath::iterator p = mp_path->begin (); p != mp_path->end (); ++p) {

    //  build the instance information from the path

    db::CplxTrans trans (p->cell_inst->complex_trans (*(p->array_inst)));
    abs_trans = abs_trans * trans;

    if (abs_coord) {
      trans = abs_trans;
    }

    std::string line;
    line += cv->layout ().cell_name (p->cell_inst->object ().cell_index ());
    line += "\tat ";
    line += trans.to_string (true /*lazy*/, dbu_coord ? 0.0 : dbu);
    
    list->addItem (line.c_str ());

  }
}

// ----------------------------------------------------------------------
//  edt::UserPropertiesForm implementation

UserPropertiesForm::UserPropertiesForm (QWidget *parent)
  : QDialog (parent)
{
  Ui::UserPropertiesForm::setupUi (this);
}

void 
UserPropertiesForm::show (lay::LayoutView *view, unsigned int cv_index, const db::Shape &shape)
{
  const lay::CellView &cv = view->cellview (cv_index);
  const db::PropertiesRepository &prep = cv->layout ().properties_repository ();

  prop_list->clear ();

  if (shape.has_prop_id ()) {

    const db::PropertiesRepository::properties_set &props = prep.properties (shape.prop_id ());
    for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
      QTreeWidgetItem *entry = new QTreeWidgetItem (prop_list);
      entry->setText (0, prep.prop_name (p->first).to_string ());
      entry->setText (1, p->second.to_string ());
    }

  }

  exec ();
}

// ----------------------------------------------------------------------
//  ObjectInstPath implementation

ObjectInstPath::ObjectInstPath ()
  : m_cv_index (0), m_topcell (0), m_layer (-1)
{
  //  .. nothing yet ..
}

db::Layout::cell_index_type 
ObjectInstPath::cell_index () const
{
  if (m_path.empty ()) {
    return m_topcell;
  } else {
    return m_path.back ().cell_inst->object ().cell_index ();
  }
}

db::CplxTrans 
ObjectInstPath::trans () const
{
  db::CplxTrans t;
  for (iterator i = m_path.begin (); i != m_path.end (); ++i) {
    t = t * i->complex_trans ();
  }
  return t;
}

void 
ObjectInstPath::remove_front (unsigned int n)
{
  while (n > 0) {
    --n;
    tl_assert (! m_path.empty ());
    if (n == 0) {
      m_topcell = m_path.front ().cell_inst->object ().cell_index ();
    }
    m_path.erase (m_path.begin ());
  }
}

void 
ObjectInstPath::insert_front (db::Layout::cell_index_type topcell, const db::InstElement &elem)
{
  tl_assert (m_topcell == elem.cell_inst->object ().cell_index ());
  m_topcell = topcell;
  m_path.insert (m_path.begin (), elem);
}

bool 
ObjectInstPath::operator< (const ObjectInstPath &d) const
{
  if (is_cell_inst () != d.is_cell_inst ()) {
    return is_cell_inst () < d.is_cell_inst ();
  }
  if (! is_cell_inst ()) {
    if (m_layer != d.m_layer) {
      return m_layer < d.m_layer;
    }
    if (m_shape != d.m_shape) {
      return m_shape < d.m_shape;
    }
  }
  if (m_cv_index != d.m_cv_index) {
    return m_cv_index < d.m_cv_index;
  }
  if (m_topcell != d.m_topcell) {
    return m_topcell < d.m_topcell;
  }
  return m_path < d.m_path;
}

bool 
ObjectInstPath::operator== (const ObjectInstPath &d) const
{
  if (is_cell_inst () != d.is_cell_inst ()) {
    return false;
  }
  if (! is_cell_inst ()) {
    if (m_layer != d.m_layer) {
      return false;
    } 
    if (m_shape != d.m_shape) {
      return false;
    }
  }
  if (m_cv_index != d.m_cv_index) {
    return false;
  }
  if (m_topcell != d.m_topcell) {
    return false;
  }
  return m_path == d.m_path;
}

}

