
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_edtUtils
#define HDR_edtUtils

#include <limits>
#include <list>
#include <utility>

#include <QtGui/QDialog>

#include "tlVector.h"
#include "dbBoxConvert.h"
#include "dbBox.h"
#include "dbLayout.h"
#include "dbShape.h"
#include "dbInstElement.h"

#include "InstantiationForm.h"
#include "UserPropertiesForm.h"

namespace db
{
  class Layout;
}

namespace lay
{
  class LayoutView;
  class Marker;
}

namespace edt {

/** 
 *  @brief A pretty heuristic method to determine the "enclosing distance" of a polygon around a point
 */
template<class Iter, class Point>
double poly_dist (Iter edge, const Point &pt)
{
  double distance = std::numeric_limits<double>::max ();

  while (! edge.at_end ()) {
    std::pair <bool, Point> ret = (*edge).projected (pt);
    if (ret.first) {
      double d (ret.second.distance (pt));
      if (d < distance) {
        distance = d;
      }
    }
    ++edge;
  }

  return distance;
}

/**
 *  @brief A generic finder class
 *
 *  A finder traverses the hierarchy and calls the virtual "find"
 *  method on each cell.
 */
class Finder
{
public:
  /**
   *  @brief Constructor
   *
   *  The point_mode is true, if the finder is supposed to operate in "point mode".
   *  In point mode, the center of the search region is the reference point. In 
   *  non-point mode, every relevant found inside the search region should be 
   *  recorded. 
   *  The base class implementation just stores this flag and provides a read
   *  accessor with the point_mode () method.
   */
  Finder (bool point_mode);

  /**
   *  @brief Destructor (just provided to please the compiler)
   */
  virtual ~Finder ();

  /**
   *  @brief Proximity getter
   *
   *  The "proximity" is the closest value passed to the "closer" method.
   *  It returns std::numeric_limits<double>::max () if nothing was found.
   */
  double proximity () const
  {
    return m_distance;
  }
  
protected:
  unsigned int layer () const
  {
    return (unsigned int) m_layer;
  }

  bool point_mode () const
  {
    return m_point_mode;
  }

  const std::vector<db::InstElement> &path () const
  {
    return m_path;
  }

  const db::Layout &layout () const
  {
    return *mp_layout;
  }

  int min_level () const
  {
    return m_min_level;
  }

  int max_level () const
  {
    return m_max_level;
  }

  bool closer (double d);

  /** 
   *  @brief Start the scan with the given parameters
   *
   *  Starts the cell scan on the given layout object, with the given region,
   *  starting at the given cell, with the given range of hierarchy levels to 
   *  consider and using just the given layer (unless -1, in which case all layers
   *  are used). For each matching cell, the "visit_cell" method is called. A 
   *  path of instantiations up to the top cell is maintained and accessible by
   *  the path() accessor.
   */
  void start (lay::LayoutView *view, unsigned int cv_index, const std::vector<db::CplxTrans> &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level, int layer = -1);

  /**
   *  @brief Provide a basic edge test facility
   *
   *  This method computes a "distance" of the edge to the reference point (the
   *  center of the search region). It updates "distance" if the computed distance
   *  is less than the one stored in "distance" or "match" is false. If the 
   *  distance is updated, match is set to true.
   */
  void test_edge (const db::DEdge &edge, double &distance, bool &match);

private:
  void do_find (const db::Cell &cell, int level, const db::CplxTrans &t);

  /**
   *  @brief Visitor sugar function
   *
   *  This method is supposed to do whatever the finder is supposed to do on the
   *  cell. It may use the "closer" method to determine if something is closer
   *  to whatever.
   */
  virtual void visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int level) = 0;

  int m_min_level, m_max_level;
  std::vector<db::InstElement> m_path;
  const db::Layout *mp_layout;
  lay::LayoutView *mp_view;
  unsigned int m_cv_index;
  db::DBox m_region;
  int m_layer;
  double m_distance;
  bool m_point_mode;
  db::box_convert <db::CellInst> m_box_convert;
  db::box_convert <db::Cell> m_cell_box_convert;
};

/**
 *  @brief A class encapsulating an instantiation path and the element addressed by it
 *
 *  This class either addresses an instance (in which case the path is that to the instance 
 *  adressed) or a shape (in which case the path leads to the cell that has this shape and
 *  the layer and shape is specified additionally).
 */

class ObjectInstPath
{
public:
  typedef std::list<db::InstElement> path_type;
  typedef path_type::const_iterator iterator;

  /**
   *  @brief Create an empty path
   */
  ObjectInstPath ();

  /**
   *  @brief Set the cellview index
   */
  void set_cv_index (unsigned int cv_index)
  {
    m_cv_index = cv_index;
  }

  /**
   *  @brief Set the topcell 
   */
  void set_topcell (db::Layout::cell_index_type topcell)
  {
    m_topcell = topcell;
  }

  /**
   *  @brief Add one element to the instantiation path
   */
  void add_path (const db::InstElement &elem)
  {
    m_path.push_back (elem);
  }

  /**
   *  @brief Add the given sequence of elements to the instantiation path
   */
  template <class Iter>
  void add_path (Iter first, Iter last)
  {
    m_path.insert (m_path.end (), first, last);
  }

  /**
   *  @brief Assign the given sequence of elements to the instantiation path
   */
  template <class Iter>
  void assign_path (Iter first, Iter last)
  {
    m_path.assign (first, last);
  }

  /**
   *  @brief Obtain the index of the target cell
   */
  db::Layout::cell_index_type cell_index () const;

  /**
   *  @brief Obtain the index of the top cell
   */
  db::Layout::cell_index_type topcell () const
  {
    return m_topcell;
  }

  /**
   *  @brief Obtain the cellview index
   */
  unsigned int cv_index () const
  {
    return m_cv_index;
  }

  /**
   *  @brief Set the layer
   *
   *  Setting a layer explicitly makes this path point to a shape (which has to be 
   *  specified additionally).
   */
  void set_layer (unsigned int layer)
  {
    m_layer = layer;
  }

  /**
   *  @brief Get the layer that the selected shape is on 
   */
  unsigned int layer () const 
  {
    tl_assert (! is_cell_inst ());
    return (unsigned int) m_layer;
  }

  /**
   *  @brief Set the selected shape
   */
  void set_shape (const db::Shape &shape)
  {
    m_shape = shape;
  }

  /**
   *  @brief Get the selected shape
   */
  const db::Shape &shape () const
  {
    tl_assert (! is_cell_inst ());
    return m_shape;
  }

  /**
   *  @brief Obtain the combined transformation for this instantiation path
   */
  db::CplxTrans trans () const;

  /**
   *  @brief Return true, if this selection represents a cell instance, not a shape
   */
  bool is_cell_inst () const
  {
    return m_layer < 0;
  }

  /**
   *  @brief Remove the given number of elements from the front of the path
   *
   *  Effectively retargets the path to a new top cell
   */
  void remove_front (unsigned int n);

  /**
   *  @brief Insert the given element at the front setting the top cell to a new one
   */
  void insert_front (db::Layout::cell_index_type topcell, const db::InstElement &elem);

  /**
   *  @brief The "begin" iterator for the path
   */
  iterator begin () const
  {
    return m_path.begin ();
  }

  /**
   *  @brief The "end" iterator for the path
   */
  iterator end () const
  {
    return m_path.end ();
  }

  /**
   *  @brief The "back" element of the path
   */
  const db::InstElement &back () const
  {
    return m_path.back ();
  }

  /** 
   *  @brief Comparison operator
   */
  bool operator< (const ObjectInstPath &d) const;

  /** 
   *  @brief Comparison operator
   */
  bool operator== (const ObjectInstPath &d) const;

  /**
   *  @brief Inequality
   */
  bool operator!= (const ObjectInstPath &d) const
  {
    return !operator== (d);
  }

private:
  unsigned int m_cv_index;
  db::Layout::cell_index_type m_topcell;
  path_type m_path;
  int m_layer;
  db::Shape m_shape;
};

/**
 *  @brief The instantiation report form
 */
class InstantiationForm 
  : public QDialog,
    public Ui::InstantiationForm
{
Q_OBJECT

public:
  InstantiationForm (QWidget *parent);
  virtual ~InstantiationForm ();

  void show (lay::LayoutView *view, const ObjectInstPath &path);

public slots:
  void display_mode_changed (bool);
  void double_clicked (QListWidgetItem *item);

private:
  void update ();

  lay::LayoutView *mp_view;
  const ObjectInstPath *mp_path;
  lay::Marker *mp_marker;
  bool m_enable_cb_callbacks;
};

/**
 *  @brief The user properties report form
 */
class UserPropertiesForm 
  : public QDialog,
    public Ui::UserPropertiesForm
{
public:
  UserPropertiesForm (QWidget *parent);

  void show (lay::LayoutView *view, unsigned int cv_index, const db::Shape &shape);
};

} // namespace edt

#endif

